// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  TU Dresden                                                            ==
// ==                                                                        ==
// ==  Institut fr Wissenschaftliches Rechnen                               ==
// ==  Zellescher Weg 12-14                                                  ==
// ==  01069 Dresden                                                         ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  https://gforge.zih.tu-dresden.de/projects/amdis/                      ==
// ==                                                                        ==
// ============================================================================

/** \file ResidualEstimator.h */

/** \defgroup Estimator Estimator module
 * @{ <img src="estimator.png"> @}
 */

#ifndef AMDIS_RESIDUALESTIMATOR_H
#define AMDIS_RESIDUALESTIMATOR_H

#include "Estimator.h"
#include "FixVec.h"

namespace AMDiS {

  /** \brief
   * Returns residual square at quadrature point. Not Member of
   * Estimator to avoid multiple instantiation.
   */
  void r(const ElInfo *elInfo,
	 int nPoints,
	 const ElementVector& uhIq,
	 const WorldVector<double> *grdUhIq,
	 const WorldMatrix<double> *D2UhIq,
	 const ElementVector& uhOldIq,
	 const WorldVector<double> *grdUhOldIq,
	 const WorldMatrix<double> *D2UhOldIq,
	 DOFMatrix *A, 
	 DOFVector<double> *fh,
	 Quadrature *quad,
	 ElementVector& result);
 
  /// Returns pow(det,2.0/dim). Not Member of Estimator to avoid multiple instantiation.
  inline double h2_from_det(double det, int dim) 
  {
    return pow(det, 2.0 / dim);
  }

  /**
   * \ingroup Estimator
   * 
   * \brief
   * Estimator for scalar problems.
   */
  class ResidualEstimator : public Estimator
  {
  public:
    /// Creator class used in the OEMSolverMap.
    class Creator : public EstimatorCreator
    {
    public:
      Creator() 
	: EstimatorCreator() 
      {}

      virtual ~Creator() 
      {}

      /// Returns a new ODirSolver object.
      Estimator* create() 
      { 
	return new ResidualEstimator(name, row);
      }
    };
  
    /// Constructor.
    ResidualEstimator(std::string name, int r);

    void init(double timestep);

    /** \brief
     * Estimates the error on an element. For more information about the parameter,
     * see the description \ref Estimator::estimateElement.
     */
    void estimateElement(ElInfo *elInfo, DualElInfo *dualElInfo = NULL);

    void exit(bool output = true);

  protected:
    /// Computes the element residual for a given element.
    double computeElementResidual(ElInfo *elInfo, DualElInfo *dualElInfo);

    /// Computes the jump residual for a given element.
    double computeJumpResidual(ElInfo *elInfo, DualElInfo *dualElInfo);

  protected:
    /// Constant in front of element residual
    double C0;
 
    /// Constant in front of edge/face residual
    double C1;

    /// Not used! Was thought to be the constant in front of coarsening term.
    double C2;

    /// Constant in front of the time
    double C3;

    /** \brief
     * Is true, if C1 != 0, and C0 and C3 = 0, hence only the jump residual must be
     * calculated. In this case, some optimizations can be done, because the jump
     * residual is calculated only on second order terms.
     */
    bool jumpResidualOnly;

    /// Number of systems, e.g., number of variables in the equation.
    int nSystems;

    /// Number of quadrature points.
    int nPoints;

    int dim;

    int degree;

    Quadrature *quad;

    FastQuadrature **quadFast;

    const BasisFunction **basFcts;

    std::vector<ElementVector> uhEl;

    std::vector<ElementVector> uhOldEl;

    std::vector<ElementVector> uhNeigh;

    ElementVector uhQP;

    ElementVector uhOldQP;

    /// Stores the element residual computed at the quadrature points of the element.
    ElementVector riq;

    WorldVector<double> *grdUh_qp;

    WorldMatrix<double> *D2uhqp; 

    ElInfo *neighInfo;

    Quadrature *surfaceQuad;

    int nPointsSurface;

    std::vector<WorldVector<double> > grdUhEl;

    std::vector<WorldVector<double> > grdUhNeigh;

    std::vector<WorldVector<double> > jump;

    std::vector<WorldVector<double> > localJump;

    WorldVector<int> faceIndEl;
    
    WorldVector<int> faceIndNeigh;

    DimVec<WorldVector<double> > *lambdaNeigh;

    DimVec<double> *lambda;

    /// Maximal number of neighbours an element may have in the used dimension.
    int nNeighbours;

    /** \brief 
     * Defines for every system if there are second order terms. These values
     * are used to ommit computations of the jump residual that is defined
     * only on second order terms.
     */    
    std::vector<bool> secondOrderTerms;
  };
}

#endif // AMDIS_RESIDUALESTIMATOR_H
